------------------------ 
--
-- Broken Loot
--
-- Usage:
--
-- /bloot
--      display the UI (WORK IN PROGRESS)
-- /bloot broadcast
--      broadcasts all the loot to the raid
local brokenLootVersion = 1.4
local enableTradeExtras = true;
local ADDONPREFIX = "BLOOT";
local _, L = ...;

local function handler(msg)
  local command, rest = msg:match("^(%S*)%s*(.-)$");
  -- Any leading non-whitespace is captured into command;
  -- the rest (minus leading whitespace) is captured into rest.
  command = strlower(command)
  if command == "" then
    --/bloot 
    displayUsageAll(true);
  elseif command == "show" or command == "s" or command == "list" then
    --/bloot show
    BLootUIShow();
  elseif command == "mode" then
    if strlower(rest) == "master" then
      handleModeChange("MASTER");
    elseif strlower(rest) == "slave" then
      handleModeChange("SLAVE");
    else
      BLootAddonSaved.Config.mode = "SLAVE";
      sendMessage(format(L["ModeSlave"]))
    end
  elseif command == "debug" then
    if strlower(rest) == "on" then
      BLootAddonSaved.Config.debugMode = 1;
      sendMessage(format(L["DebugEnabled"]))
    elseif strlower(rest) == "off" then
      BLootAddonSaved.Config.debugMode = 0;
      sendMessage(format(L["DebugDisabled"]))
    else
      sendMessage(format(L["DebugUnknownArg"]))
    end
  elseif command == "debugnospam" then
    if strlower(rest) == "on" then
      BLootAddonSaved.Config.debugMode_NoSpam = 1;
      sendMessage(format(L["DebugNoSpamEnabled"]))
    elseif strlower(rest) == "off" then
      BLootAddonSaved.Config.debugMode_NoSpam = 0;
      sendMessage(format(L["DebugNoSpamDisabled"]))
    else
      sendMessage(format(L["DebugNoSpamUnknownArg"]))
    end
  elseif BLootAddonSaved.Config.mode ~= "MASTER" then
  -- If not handled above, display some sort of help message
    displayUsageAll(true);
  else
    if command == "clear" or command == "c" then
      --/bloot clear
      actionClearList()
    elseif command == "show" or command == "s" or command == "list" then
      --/bloot show
      BLootUIShow();
      actionBroadcastList();
    elseif command == "add" or command == "a" then
        --/bloot add <who> <item>
        local lootData = {};
        lootData.who, lootData.item = rest:match("^(%S*)%s*(.-)$");
        if strlower(lootData.who) == "target" then
          lootData.who = UnitName("target")
        end
        lootData.id = generateLootId();
        lootData.when = getServerDate(true); -- now!!!
        if isValidCharacter(lootData.who) and isValidItem(lootData.item) then
          actionAddLoot(lootData);
        else
          displayUsageAdd(true);
        end
    elseif command == "roll" or command =="r" then
      actionStartRolls();
    elseif command == "stop" then
      actionStopRolls();
    elseif command == "on" then
      if strlower(rest) == "" then
        BLootAddonSaved.Config.active = 1;
        sendMessage(format(L["ActiveEnabled"]))
      else
        sendMessage(format(L["ActiveUnknownArg"]))
      end
    elseif command == "off" then
      if strlower(rest) == "" then
        BLootAddonSaved.Config.active = 0;
        sendMessage(format(L["ActiveDisabled"]))
      else
        sendMessage(format(L["ActiveUnknownArg"]))
      end
    elseif command == "countdown" then
      -- need some validation on the value
      local countdown = tonumber(rest)
      if countdown == nil then
        sendMessage(format(L["CountDownUnknownArg"], rest))
      else
        if countdown >= 5 and countdown <= 30 then
          BLootAddonSaved.Config.Countdown = countdown;
          sendMessage(format(L["CountdownSetTo"], countdown))
        else
          sendMessage(format(L["CountdownBadArg"]))
        end
      end
    else
    -- If not handled above, display some sort of help message
      displayUsageAll(true);
    end
  end
end

function displayUsageHeader(doit)
  if doit == true then
    if BLootAddonSaved.Config.mode == "MASTER" then
      sendMessage(format(L["SyntaxMaster"]));
    else
      sendMessage(format(L["SyntaxSlave"]));
    end
  end
end
function displayUsageAll(header)
  displayUsageHeader(header)
  if BLootAddonSaved.Config.mode == "MASTER" then
    displayUsageShowMaster(false);
    displayUsageClear(false);
    displayUsageAdd(false);
    displayUsageRoll(false);
  else
    displayUsageShowSlave(false);
  end
end
function displayUsageClear(header)
  displayUsageHeader(header)
  sendMessage(format(L["SyntaxClear1"]));
  sendMessage(format(L["SyntaxClear2"]));
end
function displayUsageShowMaster(header)
  displayUsageHeader(header)
  sendMessage(format(L["SyntaxShowMaster1"]));
  sendMessage(format(L["SyntaxShowMaster2"]));
end
function displayUsageShowSlave(header)
  displayUsageHeader(header)
  sendMessage(format(L["SyntaxShowSlave1"]));
  sendMessage(format(L["SyntaxShowSlave2"]));
end
function displayUsageAdd(header)
  displayUsageHeader(header)
  sendMessage(format(L["SyntaxAdd1"]));
  sendMessage(format(L["SyntaxAdd2"]));
  sendMessage(format(L["SyntaxAdd3"]));
  sendMessage(format(L["SyntaxAdd4"]));
end
function displayUsageRoll(header)
  displayUsageHeader(header)
  sendMessage(format(L["SyntaxRoll1"]));
  sendMessage(format(L["SyntaxRoll2"]));
end

---------
-- ACTIONS
function actionClearList()
  -- very simply just destroy the object for now...
  removeAllFromLootList();
  BLootAddonSaved.State = {};
  sendMessage(format(L["ClearListDone"]));
end
function actionBroadcastList()
  actionBroadcastListViaChat()
  actionBroadcastListViaAddon()
end

function actionBroadcastListViaChat()
  if BLootAddonSaved.LootList == nil then
    sendMessage(format(L["NoLootInList"]), "SELF");
  else
    local numItems = #(BLootAddonSaved.LootList) 
    if numItems > 0 then
      sendMessage(format(L["WeHaveXLoot"], numItems), "RAID"); 
      for index, lootData in ipairs(BLootAddonSaved.LootList) do
        if not(lootData.item == nil) then
          sendMessage(format(L["LootGivenBy"], lootData.item, lootData.who), "RAID");
        end
      end
    else
    sendMessage(format(L["NoLootInList"]), "SELF");
    end
  end
end

function actionBroadcastListViaAddon()
  local data = "";
  local numItems = #(BLootAddonSaved.LootList) 
  if numItems > 0 then
    for index, lootData in ipairs(BLootAddonSaved.LootList) do
      if not(lootData.item == nil) then
        data = data .. format(L["AddonChatShowBroadcastItemStructure"], lootData.who, lootData.item, lootData.id, lootData.when)
      end
    end
  end
  sendMessage(format(L["AddonChatShowBroadcast"], data), "ADDONRAID");
end

function actionAddLoot(lootData)
  if BLootAddonSaved.LootList == nil then
    BLootAddonSaved.LootList = {};
  end
  addItemToLootList(lootData)
  actionAnounceItem(lootData)
end
function actionAnounceItem(lootData)
  actionAnounceViaChat(lootData);
  actionAnounceViaAddon(lootData);
end
function actionAnounceViaChat(lootData)
  sendMessage(format(L["LootGivenBy"], lootData.item, lootData.who), "RAID");
end
function actionAnounceViaAddon(lootData)
  actionBroadcastListViaAddon()
end
function actionStartRolls()
  sendMessage("actionStartRolls()", "DEBUG");
  if BLootAddonSaved.LootList == nil then
    sendMessage(format(L["NoLootInListLeft"]), "SELF");
  else
    local numItems = #(BLootAddonSaved.LootList) 
    if numItems > 0 then
      --sendMessage(format(L["LootCount"], numItems), "RAID");
      if (BLootAddonSaved.State ~= nill) and (BLootAddonSaved.State.RollInProgress == true) and not(moreThanAMinuteAgo(BLootAddonSaved.State.RollStarted)) then
        -- we have a roll already in progress so shouldn't try to start another one.
        sendMessage(format(L["RollsInProgress"]));
      else
        BLootAddonSaved.State = {}
        sendMessage("  actionStartRolls() - starting to give it all away", "DEBUG");
        actionStartRollForItem();
      end
    else
      sendMessage(format(L["NoLootInListLeft"]), "SELF");
    end
  end
end

function actionStopRolls()
  if BLootAddonSaved.State == nill then
    sendMessage("No roll in progress to stop.", "SELF");
  else
    sendMessage("Stopping the rolls after the current one.", "RAID");
    BLootAddonSaved.State.ForceStopRolls = true;
  end
end

function actionStartRollForItem()
  sendMessage("actionStartRollForItem()", "DEBUG");
  local numItems = #(BLootAddonSaved.LootList)
  sendMessage("  actionStartRollForItem() numItems "..numItems..".", "DEBUG");
  if BLootAddonSaved.State.ForceStopRolls then
    sendMessage("Rolling of items stopped.", "RAID");
  elseif numItems > 0 then
    sendMessage("  actionStartRollForItem() finding item.", "DEBUG");
    if BLootAddonSaved.LootList[1].item == nil then
      sendMessage("Removed an item because something went wrong", "SELF");
      removeItemFromLootList(1)
    else
      BLootAddonSaved.State.RollingOn = BLootAddonSaved.LootList[1]
      BLootAddonSaved.State.RollingOn.Rolls = {}
      BLootAddonSaved.State.RollingOn.Winner = nil
      BLootAddonSaved.State.RollingOn.Unwanted = false
      BLootAddonSaved.State.Countdown = BLootAddonSaved.Config.Countdown;
      BLootAddonSaved.State.RollInProgress = true;
      BLootAddonSaved.State.RollStarted = getServerDate(true)
      --sendMessage(BLootAddonSaved.State.RollingOn.item, "DEBUG");
      sendMessage(format(L["LootCount"], numItems), "RAID");
      sendMessage(format(L["StartingRollsFor"], BLootAddonSaved.State.RollingOn.item, BLootAddonSaved.State.RollingOn.who), "RAID");
      C_Timer.After(1, actioRollForItemTick)
    end
  else
    sendMessage(L["NoLootInListLeft"], "RAID");
  end
end

function actioRollForItemTick()
  --sendMessage("actioRollForItemTick", "DEBUG");
  if BLootAddonSaved.State.RollInProgress then
    if BLootAddonSaved.State.Countdown == -1 then
      BLootAddonSaved.State.Countdown = BLootAddonSaved.State.Countdown - 1;
      C_Timer.After(1, actioRollForItemTick)
    elseif BLootAddonSaved.State.Countdown == -2 then
      actionRollForItemConfirmWinner()
    elseif BLootAddonSaved.State.Countdown == 0 then
      sendMessage("0 - Times up!", "RAID");
      BLootAddonSaved.State.Countdown = BLootAddonSaved.State.Countdown - 1;
      C_Timer.After(1, actioRollForItemTick)
    else
      if BLootAddonSaved.State.Countdown > 10 then
        if mod(BLootAddonSaved.State.Countdown, 10) == 0 then
          sendMessage(BLootAddonSaved.State.Countdown, "RAID");
        end
      elseif BLootAddonSaved.State.Countdown > 3 then
        if mod(BLootAddonSaved.State.Countdown, 5) == 0 then
          sendMessage(BLootAddonSaved.State.Countdown, "RAID");
        end
      else
        sendMessage(BLootAddonSaved.State.Countdown, "RAID");
      end
      BLootAddonSaved.State.Countdown = BLootAddonSaved.State.Countdown - 1;
      C_Timer.After(1, actioRollForItemTick)
    end
  end
end

function highestRollSort(a, b) return tonumber(a.roll) > tonumber(b.roll) end

function actionRollForItemConfirmWinner()
  local numRolls = #(BLootAddonSaved.State.RollingOn.Rolls);
  if numRolls==0 then
    sendMessage("No one wanted "..BLootAddonSaved.State.RollingOn.item..", guess it goes into the bin!", "RAID");
    BLootAddonSaved.State.RollingOn.Unwanted = true;
    removeItemFromLootList(1);
    actionStartRollForItem();
  else
    sendMessage("We have "..numRolls.." people who have rolled on "..BLootAddonSaved.State.RollingOn.item..".", "RAID");
    sort(BLootAddonSaved.State.RollingOn.Rolls, highestRollSort)
    BLootAddonSaved.State.RollingOn.Winner = BLootAddonSaved.State.RollingOn.Rolls[1];   
    actionAnnounceWinner();
    actionStartTradeLoot();
  end
end

function actionAnnounceWinner()
  actionAnnounceWinnerChat();
  actionAnnounceWinnerAddon();
end

function actionAnnounceWinnerChat()
  sendMessage("The winner of  "..BLootAddonSaved.State.RollingOn.item.." is "..BLootAddonSaved.State.RollingOn.Winner.name.." with a roll of "..BLootAddonSaved.State.RollingOn.Winner.roll, "RAID");
end

function actionAnnounceWinnerAddon()
  --sendMessage("actionAnnounceWinnerAddon - NYI");
end

function actionStartTradeLoot()
  if enableTradeExtras then
    sendMessage("actionTradeLoot - starting", "DEBUG");
    if not(alreadyTrading()) then
      if #(BLootAddonSaved.LootList) > 0 then
        lootData = BLootAddonSaved.LootList[1]
        if lootData.Winner == nil then
        -- code shouldn't get here anymore
          sendMessage("  actionTradeLoot - no one wanted! - SHOULD NEVER BE SEEN!!!!********************", "DEBUG");
          if lootData.Unwanted == true then
            sendMessage("No one wanted "..lootData.item.." so removing from list!");
            removeItemFromLootList(1)
            sendMessage("  actionTradeLoot - removed that item from the list!", "DEBUG");
            actionStartRollForItem();
          else
            sendMessage("No winner for "..lootData.item.." but not unwanted?!");
            actionStartRollForItem();
          end
        else
          sendMessage("  actionTradeLoot - giving away "..lootData.item.." to "..lootData.Winner.name..".", "DEBUG");
          --ClearTarget();
          if lootData.Winner.name == UnitName("player") then
            sendMessage("You won "..lootData.item.."!", "SELF");
            removeItemFromLootList(1);
            sendMessage("  actionTradeLoot - removed that item from the list!", "DEBUG");
            actionStartRollForItem();
          else
            inRange = CheckInteractDistance(lootData.Winner.name, 2);
            if inRange then 
              InitiateTrade(lootData.Winner.name)
              currentTrade.startedByMe = true;
              currentTrade.tradeWith = lootData.Winner.name;
              currentTrade.wantToTradeWith = lootData.Winner.name;
            else
              sendMessage(lootData.Winner.name.." is too far away from "..UnitName("player").." to trade!", "RAID");
              actionStartRollForItem();
            end
          end
        end
      end
    end
  end
end

function addItemToLootList(lootData, silent)
  tinsert(BLootAddonSaved.LootList, lootData);
  if silent ~= true then
    BLLootLootFrameEntry_Update();
    actionBroadcastListViaAddon();
  end
end

function removeItemFromLootList(id, silent)
  tremove(BLootAddonSaved.LootList, id);
  if silent ~= true then
    BLLootLootFrameEntry_Update();
    actionBroadcastListViaAddon();
  end
end
function removeAllFromLootList(silent)
  BLootAddonSaved.LootList = {};
  if silent ~= true then
    BLLootLootFrameEntry_Update();
    actionBroadcastListViaAddon();
  end
end

---------
-- HANDLERS
function handleModeChange(mode)
  if mode == "MASTER" then
    BLootAddonSaved.Config.mode = "MASTER";
    BLRollAllButton:Show();
    BLClearAllButton:Show();
    sendMessage(format(L["ModeMaster"]))
  else
    BLootAddonSaved.Config.mode = "SLAVE";
    BLRollAllButton:Hide();
    BLClearAllButton:Hide();
    sendMessage(format(L["ModeSlave"]))
  end
end

function handleRoll(name, roll, minRoll, maxRoll)
  if BLootAddonSaved.State.RollInProgress then
    -- Should we really care?
    if BLootAddonSaved.State.Countdown > 0 then
      local found = false;
      for index, roll in ipairs(BLootAddonSaved.State.RollingOn.Rolls) do
        if roll.name == name then
          found = true;
        end
      end
      if found == true then
        sendMessage(name.." is a dirty rotton cheat rolling more than once :)", "RAID")
      elseif minRoll == "1" and maxRoll == "100" then
        sendMessage(name.." rolled "..roll..".", "RAID")
        tinsert(BLootAddonSaved.State.RollingOn.Rolls, {name = name, roll = roll})
      else
        sendMessage(name.." is a cheat on rolls!!!!!"..minRoll.."/"..maxRoll, "RAID")
      end
    end
  end
end

function handleRegisterInterest(id, name, item)

end
      
function handleTradeComplete()
  local finishedTrade = copyTrade(currentTrade);
  currentTrade = resetTrade();
  if finishedTrade.cancel == 0 then  
    finishedTrade.openTrade = false;
    -- did we complete a trade?
    sendMessage("handleTradeComplete()", "DEBUG")
    local numTargetItems =  #(finishedTrade.targetItems)
    sendMessage("  handleTradeComplete() - Recieved "..numTargetItems.." items", "DEBUG")
    if numTargetItems > 0 then
      -- recieved an item or two I think....?
        for i = 1, numTargetItems do
          --sendMessage("  handleTradeComplete() - adding "..finishedTrade.targetItems[i]..".", "DEBUG")
          local lootData = {};
          lootData.who = finishedTrade.tradeWith; 
          lootData.item = finishedTrade.targetItems[i];
          lootData.id = generateLootId();
          lootData.when = getServerDate(true); -- now!!!
          actionAddLoot(lootData);
        end
    end
    local numPlayerItems =  #(finishedTrade.playerItems)
    sendMessage("  handleTradeComplete() - Given away "..numPlayerItems.." items", "DEBUG")
    if numPlayerItems > 0 then
      -- need to integrate the queue here to make sure we are giving items to who needs them
      for i = 1, numPlayerItems do
        sendMessage("  handleTradeComplete() - given away "..finishedTrade.playerItems[i].." to "..finishedTrade.tradeWith..".", "DEBUG")
        for lootIndex, lootData in ipairs(BLootAddonSaved.LootList) do
          if lootData.item == finishedTrade.playerItems[i] then
            removeItemFromLootList(lootIndex)
            sendMessage("  handleTradeComplete() - have given away "..lootData.item..".", "DEBUG")
          else
            sendMessage("  handleTradeComplete() - "..finishedTrade.playerItems[i].." is NOT the same as "..lootData.item..".", "DEBUG")
          end
        end
      end
      if #(BLootAddonSaved.LootList)>0 then
        actionStartRollForItem();
      end
    end
  else
    sendMessage("  handleTradeComplete() - cancelled trade?", "DEBUG")
  end
end

function handleTradeStarted()
  currentTrade.tradeWith = UnitName("NPC")
  if currentTrade.startedByMe == true then
    sendMessage("  TRADE_SHOW() I started it (I think)", "DEBUG")
    if #(BLootAddonSaved.LootList) > 0 then
      lootData = BLootAddonSaved.LootList[1]
      if lootData.Winner.name == currentTrade.tradeWith then
        sendMessage("  TRADE_SHOW() - wanting to give "..lootData.item.." to "..lootData.Winner.name..".", "DEBUG")
        --Should actually give the item away here....
        local bag, slot = findItemInBags(lootData.item)
        if bag == nil or slot == nil then
          sendMessage("  TRADE_SHOW() - I don't have the item to give away?", "DEBUG")
        else
          UseContainerItem(bag, slot)
          -- not auto accepting that is too scary
        end
      else
        sendMessage("  TRADE_SHOW() - Think I started it but the name i have doesn't match the person recieving the loot?", "DEBUG")
      end 
    else
      sendMessage("  TRADE_SHOW() - Think I started it but have no loot to give away?", "DEBUG")
    end
  else
    sendMessage("  TRADE_SHOW() probably not started by me", "DEBUG")
  end
end

local previousAddonChatMessage = ""
function handleAddonChat(prefix, message, channel, sender)
  local playerAndRealm = "FISH" --format("%s-%s", GetUnitName("player", true),  GetRealmName())
  if sender~=playerAndRealm then
    -- check what type it was
    local msgType = strsub(message, 1, 5)
    if msgType == L["AddonChatChunkPrefix"] then
      --sendMessage(format("AddonChatChunk: %s", message));
      local chunkN, maxChunk, data = message:match(L["AddonChatChunkRegEx"])
      if chunkN == nil then chunkN = 999 end
      if maxChunk == nil then maxChunk = 999 end
      if data == nil then data = "nil" end
      if chunkN == 1 then
        previousAddonChatMessage = data;
      else
        previousAddonChatMessage = previousAddonChatMessage..data;
        if chunkN == maxChunk then
          --sendMessage(format("Final unchunked data %q", previousAddonChatMessage), "DEBUG")
          handleAddonChat(prefix, previousAddonChatMessage, channel, sender);
        end
      end
    elseif msgType == L["AddonChatStartupBroadcastPrefix"] then
      local version = message:match(L["AddonChatStartupBroadcastRegEx"])
      if version ~=null then
        sendMessage(format("  handleAddonChat - version %q", version), "DEBUG")
        version = tonumber(version)
        if version == brokenLootVersion then
          sendMessage(format("%s is running the same version of Broken's Loot Addon (%s)", sender, version), "DEBUG")
        elseif version > brokenLootVersion then
          sendMessage(format("There is a newer version of Broken's Loot Addon available"), "SELF")
        else
          sendMessage(format("%s is running an older version of Broken's Loot Addon (%s)", sender, version), "DEBUG")
        end
      else
        sendMessage(format("  handleAddonChat - version NULL"), "DEBUG")
      end
    elseif msgType == L["AddonChatShowBroadcastPrefix"] then
      local data = strsub(message, 6);
      --local version = message:match(L["AddonChatStartupBroadcastRegEx"])
        handleNewListDataRecieved(data);
        sendMessage("Recieving the data still to be sorted, but here is the UI at least!!", "SELF")  
        BLootUIShow();
    else
      sendMessage(format("  handleAddonChat - unknown msgType %q", msgType), "DEBUG")
    end
  else
    sendMessage("Recieved data from myself", "DEBUG")
  end
end

function handleNewListDataRecieved(data)
  --sendMessage(format("handleNewListDataRecieved(%q)", data), "DEBUG")
  if data == "*" then
    sendMessage(format("  handleNewListDataRecieved NO DATA so clearing", data), "DEBUG")
    -- need to clear out my local copy of the data here....
    removeAllFromLootList(true);
    BLLootLootFrameEntry_Update();
  else
    removeAllFromLootList(true);
    local c = 0; -- just in here to kill the loop in case of problems :) 
    while (data ~= nil) and (data ~= "") do
      local who, item, id, when, rest = data:match(L["AddonChatShowBroadcastItemRegEx"])
      data = rest;
      local lootData = {};
      lootData.who = who;
      lootData.item = item;
      lootData.id = id;
      lootData.when = when;
      addItemToLootList(lootData, true);
      --sendMessage("Adding item to list", "DEBUG")
      c = c +1;
      if c > 1000 then 
        break; 
      end;
    end
    BLLootLootFrameEntry_Update();
  end

end

---------
-- UTILITY FUNCTIONS
function sendMessage(message, channel)
  if message == nil then
    message = "";
  end;
  if type(message) == "table" then
    local tableMessage = message
    message = "["
    for index,value in pairs(tableMessage) do 
      if type(value) == "table" then
        local tableMessage2 = value
        message = message.."["
        for index2,value2 in pairs(tableMessage2) do 
          if type(value2) == "table" then
            message = message..tostring(index2).." : [another table], "
          else
            message = message..tostring(index2).." : "..value2..", "
          end
        end
        message = message.."]"
      else
        message = message..tostring(index).." : "..value..", "
      end
    end
    message = message.."]"
  end
  if channel == nil then
    channel = "SELF";
  end;
  if BLootAddonSaved.Config.debugMode_NoSpam == 1 then
    DEFAULT_CHAT_FRAME:AddMessage("Broken's Loot ("..channel.."): "..message);
  else
    if channel=="RAID" then
      if GetNumGroupMembers() > 5 then
        SendChatMessage("Broken's Loot: "..message, "RAID");
      elseif GetNumGroupMembers() > 0 then
        SendChatMessage("Broken's Loot: "..message, "PARTY");
      else 
        DEFAULT_CHAT_FRAME:AddMessage("Broken's Loot: "..message);
      end
    elseif (channel=="GUILD") then
        SendChatMessage("Broken's Loot: "..message, "GUILD") 
    elseif (channel=="SELF") then
      DEFAULT_CHAT_FRAME:AddMessage("Broken's Loot: "..message);
    elseif (channel=="DEBUG") and (BLootAddonSaved.Config.debugMode == 1) then
      DEFAULT_CHAT_FRAME:AddMessage("Broken's Loot (DEBUG): "..message, 1, 1, 0);
    elseif (channel=="ADDONRAID") then
      -- need to chop the message into 250 character chunks....
      local messageLen = strlen(message);
      if messageLen > 250 then
        -- split it into 200 character chunks here
        local maxChunks = ceil(messageLen/240);
        --sendMessage(format("SendMessage: Addon chat message too long so chunking %d = %d chunks", messageLen, maxChunks), "DEBUG");
        for chunkN = 1, maxChunks do
          local chunkMessage = strsub(message, 1, 240)
          message = strsub(message, 241)
          local toSend = format(L["AddonChatChunk"], chunkN, maxChunks, chunkMessage);
          SendAddonMessage(ADDONPREFIX, toSend, "GUILD", "")
        end
      else
        SendAddonMessage(ADDONPREFIX, message, "GUILD", "")      
      end
      --DEFAULT_CHAT_FRAME:AddMessage("SendAddonMessage: "..message, 1, 1, 0);
    end
  end
end

function isValidCharacter(characterName)
  if characterName == "Someone" then
    return true;
  end
  local i = getCharacterGroupIndex(characterName);
  sendMessage("isValidCharacter('"..characterName.."') is group at: "..i..".", "DEBUG");
  if i > 0 then
    return true
  else
    sendMessage("isValidCharacter('"..characterName.."') is not in the group", "DEBUG");
    local numGroupMembers = GetNumGroupMembers()
    if numGroupMembers == 0 then
      local name = GetUnitName("player", true)
      if strlower(name) == strlower(characterName) then
        sendMessage("isValidCharacter('"..characterName.."') is the current character", "DEBUG");
        return true
      end
    end
  end
  return false
end

function getCharacterGroupIndex(characterName)
  sendMessage("getCharacterGroupIndex('"..characterName.."')", "DEBUG");
  local numGroupMembers = GetNumGroupMembers()
  for ci = 1, numGroupMembers do
    if strlower(GetRaidRosterInfo(ci)) == strlower(characterName) then
      sendMessage("getCharacterGroupIndex('"..characterName.."') = "..ci, "DEBUG");
      return ci;
    end
  end
  sendMessage("getCharacterGroupIndex('"..characterName.."') = not in group!", "DEBUG");
  return 0;
end

function isValidItem(itemLink)
  -- not checking it is in my bags initially so that I don't cause issues...
  --if findItemInBags(itemLink)
  --sendMessage("isValidItem - NYI");
  --debugItem(itemLink)
  return true;
end

function debugItem(itemLink)
  sendMessage("debugItem("..itemLink..")", "DEBUG");
  local itemName, itemLink, itemRarity, itemLevel, itemMinLevel, itemType, itemSubType, itemStackCount, itemEquipLoc, itemTexture, itemSellPrice = GetItemInfo(itemLink) 
  sendMessage("  itemType:"..itemType, "DEBUG");
  sendMessage("  itemSubType:"..itemSubType, "DEBUG");
  sendMessage("  itemEquipLoc"..itemEquipLoc, "DEBUG");
  getItemStats(itemLink);
end;

function getItemStats(itemLink)
  local stats = {};
  GetItemStats(itemLink, stats);
  sendMessage("getItemStats("..itemLink..")", "DEBUG")
  for key, value in pairs(stats) do 
    sendMessage("  "..key.." - " ..value, "DEBUG")
  end
  local agility, intellect, strength, stamina;
  agility = (stats["ITEM_MOD_AGILITY_SHORT"] or 0);
  intellect = (stats["ITEM_MOD_INTELLECT_SHORT"] or 0);
  strength = (stats["ITEM_MOD_STRENGTH_SHORT"] or 0);
  stamina = (stats["ITEM_MOD_STAMINA_SHORT"] or 0);
  sendMessage("  agi "..agility.." int "..intellect.." str "..strength.." stam "..stamina, "DEBUG");
  return agility, intellect, strength, stamina;
end

function generateLootId()
  BLootAddonSaved.lootId = BLootAddonSaved.lootId + 1;
  sendMessage("generateLootId() = "..BLootAddonSaved.lootId, "DEBUG")
  return BLootAddonSaved.lootId;
end

function getServerDate(andTime)
  local weekday, month, day, year = CalendarGetDate();
  if month < 10 then month = "0"..month; end;
  if day < 10 then day = "0"..day; end;
  if andTime then
    local hours,minutes = GetGameTime();
    if hours < 10 then hours = "0"..hours; end;
    if minutes < 10 then minutes = "0"..minutes; end;
    local currentDateTime = day.."/"..month.."/"..year.." "..hours..":"..minutes..":00";
    return currentDateTime;
  else
    local currentDate = day.."/"..month.."/"..year;
    return currentDate;
  end;
end

function moreThanAMinuteAgo(when)
-- NYI
  return true;
end

function findItemInBags(findItemId)
  --sendMessage("findItemInBags("..findItemId..")", "DEBUG");
  for bag = 0, 5, 1 do 
    --sendMessage("  Looking in bag ("..bag..")", "DEBUG");
    local numberOfSlots = GetContainerNumSlots(bag);
    --sendMessage("  Bag "..bag.." has "..numberOfSlots.." slots.", "DEBUG");
    for slot = 1, numberOfSlots+1, 1 do
      itemLink = GetContainerItemLink(bag, slot);
      if not(itemLink == nil) then
        --sendMessage("  Found at ("..bag..")("..slot..")"..itemLink, "DEBUG");
        if findItemId == itemLink then
          --sendMessage("  Found at ("..bag..")("..slot..")", "DEBUG");
          return bag, slot;
        end
      end
    end
  end
  --sendMessage("  Item not found", "DEBUG");
  return nil, nil;
end

function alreadyTrading()
  if enableTradeExtras then
    return currentTrade.openTrade;
  else
    return true;
  end
end

--[[ Function: extractItemLinkFromChatItemLink
     Purpose: gets an item link string from a chat item link string
     Note: don't ask questions, it works
  ]]
function extractItemLinkFromChatItemLink(message)
  local _, _, _, _itemLink = string.find(message, "(|H(.+)|h)");
  _, _, _itemLink = string.find(_itemLink, "(.-|h)");
  return _itemLink;
end

function shallowcopy(orig)
    local orig_type = type(orig)
    local copy
    if orig_type == 'table' then
        copy = {}
        for orig_key, orig_value in pairs(orig) do
            copy[orig_key] = orig_value
        end
    else -- number, string, boolean, etc
        copy = orig
    end
    return copy
end
---------
-- STARTUP
SLASH_BROKENSLOOT1 = '/bloot';
SlashCmdList["BROKENSLOOT"] = handler;

function startup()
  -- brokenLootVersion
  local defaultConfig = {
    debugMode = 0,
    debugMode_NoSpam = 0,
    Countdown = 25,
    active = 0,
    mode = "SLAVE"
  }; 
  if BLootAddonSaved ~= nil and BLootAddonSaved.versionId ~= brokenLootVersion then
    -- been here before but not the same version as before so reset time.
    defaultConfig.debugMode = BLootAddonSaved.Config.debugMode;
    defaultConfig.debugMode_NoSpam = BLootAddonSaved.Config.debugMode_NoSpam;
    defaultConfig.Countdown = BLootAddonSaved.Config.Countdown;
    defaultConfig.active = BLootAddonSaved.Config.active;
    defaultConfig.mode = BLootAddonSaved.Config.mode;
    BLootAddonSaved = nil;
  end
  if BLootAddonSaved == nil then
    BLootAddonSaved = {
      lootId = 0,
      versionId = brokenLootVersion,
      Config = defaultConfig
    }
  end
  if BLootAddonSaved.Config.active == 1 then 
    sendMessage("Loaded and active!", "SELF") 
    sendMessage("Countdown: "..BLootAddonSaved.Config.Countdown, "SELF") 
    if BLootAddonSaved.Config.debugMode == 1 then 
      sendMessage("Debug: on", "SELF") 
    end;
    if BLootAddonSaved.Config.debugMode_NoSpam ==  1 then 
      sendMessage("Debug No Spam: on", "SELF") 
    end;
  else
    sendMessage("Loaded but deactivated.");
  end;
  if BLootAddonSaved.Config.mode ==  "MASTER" then 
    sendMessage("Running in Master mode", "SELF") 
  end;
  local ok = RegisterAddonMessagePrefix(ADDONPREFIX)
  --sendMessage(format("  RegisterAddonMessagePrefix(BLOOT) = []", ok), "DEBUG");
  --sendMessage(format(L["AddonChatStartupBroadcast"], brokenLootVersion), "ADDONRAID"); 
end

---------
-- ADDON VARIABLES
BrokensLoot = {}; -- holder for the panels

---------
-- EVENTS
local frame, events = CreateFrame("Frame"), {};

function events:CHAT_MSG_SYSTEM (...)
  --sendMessage("CHAT_MSG_SYSTEM", "DEBUG")
  local arg1 = select(1,...)
  if arg1 then
    local name, roll, minRoll, maxRoll = arg1:match(L["RandomRollResultRegEx"])
    if name then
      -- we have found a roll - I think :)
      handleRoll(name, roll, minRoll, maxRoll);
    end
  end
end

function resetTrade()
  local t = {};
  t.openTrade = false;
  t.tradeWith = "Someone";
  t.startedByMe = false;
  t.wantToTradeWith = "";
  t.acceptPlayer = 0;
  t.acceptTarget = 1;
  t.playerItems = {};
  t.targetItems = {};
  t.close = 0;
  t.cancel = 0;
  return t;
end
function copyTrade(e)
  local t = {};
  t.openTrade = e.openTrade;
  t.tradeWith = e.tradeWith;
  t.startedByMe = e.startedByMe;
  t.wantToTradeWith = e.wantToTradeWith;
  t.acceptPlayer = e.acceptPlayer;
  t.acceptTarget = e.acceptTarget;
  t.playerItems = e.playerItems;
  t.targetItems = e.targetItems;
  t.close = e.close;
  t.cancel = e.cancel;
  return t;
end

currentTrade = resetTrade();

if enableTradeExtras then
  function events:TRADE_SHOW(...)
    if BLootAddonSaved.Config.active then
      sendMessage("TRADE_SHOW()", "DEBUG")
      currentTrade.openTrade = true;
      handleTradeStarted()
    end
  end

  function events:TRADE_ACCEPT_UPDATE(...)
    if BLootAddonSaved.Config.active then
      local arg1, arg2  = select(1,...)
      currentTrade.acceptPlayer = arg1
      currentTrade.acceptTarget = arg2
    end
  end
  -- don't think this event exists any more
  function events:TRADE_REQUEST(...)
    if BLootAddonSaved.Config.active then
      sendMessage("TRADE_REQUEST()", "DEBUG")
      local arg1  = select(1,...)
      currentTrade.tradeWith = arg1;
    end
  end

  -- player cancelled - CLOSE CLOSE CANCEL  ??
  -- target cancelled - CANCEL CLOSE CLOSE  ??

  -- Fired when a trade attempt is cancelled. Fired after TRADE_CLOSE when aborted by player, before TRADE_CLOSE when done by target.
  -- Upon a trade being cancelled (as in, either part clicking the cancel button), TRADE_CLOSE is fired twice, and then TRADE_REQUEST_CANCEL once.
  function events:TRADE_REQUEST_CANCEL(...)
    if BLootAddonSaved.Config.active then
      sendMessage("TRADE_REQUEST_CANCEL()", "DEBUG")
      currentTrade.cancel = currentTrade.cancel + 1;
      if currentTrade.inProgress == true then
      else
        if currentTrade.inProgress == true then
        else
          currentTrade.inProgress = true;
          C_Timer.After(2, handleTradeComplete)
        end
      end
    end
  end

  function events:TRADE_CLOSED(...)
    if BLootAddonSaved.Config.active then
      -- dirty hack in here for now - you have to target the person before hitting accept so i know who it is? just a test
      --currentTrade.tradeWith = UnitName("target")
      currentTrade.close = currentTrade.close + 1;
      sendMessage("TRADE_CLOSED()", "DEBUG")
      if currentTrade.inProgress == true then
      else
        if currentTrade.inProgress == true then
        else
          currentTrade.inProgress = true;
          C_Timer.After(2, handleTradeComplete)
        end
      end
    end
  end

  function events:TRADE_TARGET_ITEM_CHANGED(...)
    if BLootAddonSaved.Config.active then
      local arg1 = select(1,...)
      sendMessage("TRADE_PLAYER_ITEM_CHANGED("..arg1..")", "DEBUG")
      currentTrade.playerItems = {};
      currentTrade.targetItems = {};
      for tradeSlotId = 1, 7 do
        playerSideItemLink = GetTradePlayerItemLink(tradeSlotId);
        if not(playerSideItemLink == nil) then
          sendMessage("  playerSideItemLink: "..playerSideItemLink..".", "DEBUG")
          tinsert(currentTrade.playerItems, playerSideItemLink);
        end
        targetSideItemLink = GetTradeTargetItemLink(tradeSlotId);
        if not(targetSideItemLink == nil) then
          sendMessage("  targetSideItemLink"..targetSideItemLink..".", "DEBUG")
          tinsert(currentTrade.targetItems, targetSideItemLink);
        end
      end
    end
  end
  function events:TRADE_PLAYER_ITEM_CHANGED(...)
    if BLootAddonSaved.Config.active then
      local arg1 = select(1,...)
      sendMessage("TRADE_PLAYER_ITEM_CHANGED("..arg1..")", "DEBUG")
      currentTrade.playerItems = {};
      currentTrade.targetItems = {};
      for tradeSlotId = 1, 7 do
        playerSideItemLink = GetTradePlayerItemLink(tradeSlotId);
        if not(playerSideItemLink == nil) then
          sendMessage("  playerSideItemLink: "..playerSideItemLink..".", "DEBUG")
          tinsert(currentTrade.playerItems, playerSideItemLink);
        end
        targetSideItemLink = GetTradeTargetItemLink(tradeSlotId);
        if not(targetSideItemLink == nil) then
          sendMessage("  targetSideItemLink"..targetSideItemLink..".", "DEBUG")
          tinsert(currentTrade.targetItems, targetSideItemLink);
        end
      end
    end
  end
end

function events:CHAT_MSG_ADDON(...)
  if BLootAddonSaved.Config.active then
    local arg1, arg2, arg3, arg4 = select(1,...)
    if(arg1==ADDONPREFIX) then
      -- filter out everything other than stuff i care about...
      --sendMessage("CHAT_MSG_ADDON()", "DEBUG")
      --sendMessage(format("  arg1[%s] arg2[%s] arg3[%s] arg4[%s]", arg1, arg2, arg3, arg4), "DEBUG")
      handleAddonChat(arg1, arg2, arg3, arg4);
    end
  end
end


function events:VARIABLES_LOADED (...)
  startup();
end

frame:SetScript("OnEvent", function(self, event, ...)
  events[event](self, ...); -- call one of the functions above
end);

for k, v in pairs(events) do
  frame:RegisterEvent(k); -- Register all events for which handlers have been defined
end


---------
-- UI STUFF

function BLootUIShow()
  if BLootAddonSaved.Config.mode == "MASTER" then
    BLRollAllButton:Show();
    BLClearAllButton:Show();
  else
    BLRollAllButton:Hide();
    BLClearAllButton:Hide();
  end
  BLootMainDisplayFrame:Show();
end
function BLLootLootFrameEntry_Update()
  BLLootLootScrollBar_Update();
end

function BLLootLootScrollBar_Update()
  local maxRows = 8;
  local loot = BLootAddonSaved.LootList;
  local lootCount = 0;
  if loot ~= nil then
    lootCount = #(loot);
    --sendMessage(format("BLLootLootScrollBar_Update: lootCount: %i", lootCount), "DEBUG")
  end
  local line; -- 1 through 5 of our window to scroll
  local lineplusoffset; -- an index into our data calculated from the scroll offset
  FauxScrollFrame_Update(BLLootLootScrollBar, lootCount, maxRows, 16);
  for line=1,maxRows do
    lineplusoffset = line + FauxScrollFrame_GetOffset(BLLootLootScrollBar);
    local button = getglobal("BLLootLootFrameEntry"..line);
    local buttonClear = getglobal("BLLootLootFrameAction"..line);
    button:UnlockHighlight();
    if lineplusoffset <= lootCount then
      --sendMessage(format("BLLootLootScrollBar_Update: lineplusoffset: %i", lineplusoffset), "DEBUG")
      button:SetText(format(L["AddonLootLine"], loot[lineplusoffset].item, loot[lineplusoffset].who));
      button:Show();
      if buttonClear ~= nil then buttonClear:Show(); end
--      if loot[lineplusoffset]==BrokenLootSettings.LootAssign.loot then
--        button:LockHighlight();
--      end  
    else
        button:Hide();
        if buttonClear ~= nil then buttonClear:Hide(); end
    end
  end
end

function BLLootFrameEntry_Update(self)
  BLLootLootFrameEntry_Update();
end

function BLClearAllButton_OnClick(self)
-- ask for confirmation before doing it!!!
  actionClearList()
end

function BLRollAllButton_OnClick(self)
-- ask for confirmation before doing it!!!
  actionStartRolls()
end

function BLLootLootFrameAction_Click(self)
  local line = self:GetName();
  line = strsub(line, 22)
  lineplusoffset = line + FauxScrollFrame_GetOffset(BLLootLootScrollBar);
  removeItemFromLootList(lineplusoffset)
  BLLootLootFrameEntry_Update();
end